from brigid.library.tests import make as library_make
from brigid.validation.page_validators import page_is_rendered


class TestPageIsRendered:

    def test_empty_body(self) -> None:
        page = library_make.page(body="")
        assert page_is_rendered(page) == []

    def test_no_errors(self) -> None:
        page = library_make.page(body="some text")
        assert page_is_rendered(page) == []

    def test_has_errors(self) -> None:
        body = """
/// brigid-images
x = "y"
///
"""

        page = library_make.page(body=body)
        assert page_is_rendered(page) != []



import warnings

from bs4 import BeautifulSoup, MarkupResemblesLocatorWarning


def strip_html(html: str) -> str:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=MarkupResemblesLocatorWarning)
        soup = BeautifulSoup(html, "html.parser")
        return soup.get_text()



from brigid.library.connectivity import connectivity
from brigid.library.entities import Page, PageSimilarityScore
from brigid.library.storage import storage


# TODO: collect data to explain similarity
# TODO: add similarity by vector embedings?
def get_similar_pages(language: str, original_page: Page, number: int) -> list[PageSimilarityScore]:  # noqa: CCR001

    site = storage.get_site()

    scores = []

    # initialize scores
    for page in storage.last_pages(language=language):
        if original_page.id == page.id:
            continue

        scores.append(PageSimilarityScore(page_id=page.id, score=0))

    # add points for common tags
    for page_score in scores:
        page = storage.get_page(id=page_score.page_id)

        for tag in original_page.tags:
            if tag in site.similarity.ignore_similar_tags:
                continue

            if tag in page.tags:
                page_score.add_score(site.similarity.common_tag_score, f'common tag "{tag}"')

    # add points for special tags
    for page_score in scores:
        page = storage.get_page(id=page_score.page_id)

        for tag in page.tags:
            if tag in site.similarity.bonus_for_tags:
                page_score.add_score(site.similarity.bonus_for_tags[tag], f'special tag "{tag}"')

    # add points for linked articles

    for page_score in scores:
        references_number = len(connectivity.get_referenced_from(page_id=page_score.page_id))
        page_score.add_score(
            references_number * site.similarity.bonus_per_reference,
            f"bonus for {references_number} references",
        )

    for referenced_to_id in connectivity.get_referenced_to(page_id=original_page.id):
        # TODO: optimize to map
        for page_score in scores:
            if page_score.page_id == referenced_to_id:
                page_score.add_score(
                    site.similarity.referenced_to_score,
                    f"referenced to {referenced_to_id}",
                )

    for referenced_from_id in connectivity.get_referenced_from(page_id=original_page.id):
        # TODO: optimize to map
        for page_score in scores:
            if page_score.page_id == referenced_from_id:
                page_score.add_score(
                    site.similarity.referenced_from_score,
                    f"referenced from {referenced_from_id}",
                )

    # selecting top pages
    scores.sort(key=lambda x: x.score, reverse=True)

    return scores[:number]



import re
import xml.etree.ElementTree as etree  # noqa: S405

from markdown.extensions import Extension
from markdown.treeprocessors import Treeprocessor

from brigid.domain.urls import UrlsPost
from brigid.renderer.context import render_context

DISALLOWED_CHARS_PATTERN = re.compile(r"[^-a-zA-Z0-9]+")


class AnchorCounter:

    def __init__(self):
        self.counts = [0, 0, 0, 0, 0]

    def count(self, tag: str) -> None:
        if "h1" == tag:
            raise NotImplementedError("We do not expect h1 here")

        index = int(tag[1]) - 2

        self.counts[index] += 1

        for i in range(index + 1, len(self.counts)):
            self.counts[i] = 0

    def anchor(self) -> str:
        counts = list(self.counts)

        while counts[-1] == 0:
            counts.pop()

        return f'h-{"-".join(str(c) for c in counts)}'


class HeaderAnchorsProcessor(Treeprocessor):

    def run(self, root: etree.Element) -> None:
        context = render_context.get()

        counter = AnchorCounter()

        headers = [tag for tag in root.iter() if tag.tag in ("h1", "h2", "h3", "h4", "h5", "h6")]

        for header in headers:

            if header.tag == "h1":
                context.add_error(
                    failed_text=str(header),
                    message="H1 headers are not allowed in articles, they are for titles only",
                )
                continue

            counter.count(header.tag)

            if header.text is None:
                text = None
            else:
                text = header.text.strip()

            id = counter.anchor()

            post_url = UrlsPost(language=context.page.language, slug=context.article.slug)

            url = f"{post_url.url()}#{id}"

            header.set("id", id)

            anchor = etree.Element("a")
            anchor.set("href", url)
            anchor.text = text

            header.text = None
            header.insert(0, anchor)


class HeaderAnchorsExtension(Extension):

    def extendMarkdown(self, md):
        # TODO: which priority to set?
        md.treeprocessors.register(HeaderAnchorsProcessor(), "brigid_header_anchors", 175)



import contextlib
import contextvars
import pathlib
from typing import Generator

import pydantic
from sentry_sdk import capture_message

from brigid.core.entities import BaseEntity
from brigid.library.entities import Article, Page


class RenderError(BaseEntity):
    filepath: pathlib.Path | None = None
    failed_text: str
    message: str


class RenderContext(BaseEntity):
    page: Page
    article: Article
    content: str | None = None
    renderer: int
    errors: list[RenderError] = pydantic.Field(default_factory=list)

    def add_error(self, failed_text: str, message: str) -> None:
        self.errors.append(RenderError(failed_text=failed_text, message=message, filepath=self.page.path))

    model_config = pydantic.ConfigDict(frozen=False)


render_context: contextvars.ContextVar[RenderContext] = contextvars.ContextVar("brigid_markdown_context")


@contextlib.contextmanager
def markdown_context(context: RenderContext) -> Generator[None, None, None]:

    token = render_context.set(context)

    try:
        yield
    finally:
        render_context.reset(token)

        prev_context = render_context.get(None)

        if prev_context is not None:
            prev_context.errors.extend(context.errors)

        if prev_context is None and context.errors:
            capture_message(repr(context.errors[0]))



from typing import Iterable

import fastapi

from brigid.api.default_translations import translations


def parse_accept_language(accept_language):
    languages = accept_language.split(",")

    parsed_languages = []

    for language in languages:
        parts = language.split(";q=")

        if len(parts) == 1:
            parsed_languages.append((parts[0].strip(), 1.0))

        else:
            parsed_languages.append((parts[0].strip(), float(parts[1])))

    parsed_languages.sort(key=lambda x: x[1], reverse=True)

    return [lang.lower() for lang, _ in parsed_languages]


def select_language(accepted_languages, available_languages):
    for lang in accepted_languages:
        if lang in available_languages:
            return lang

        if "-" not in lang:
            continue

        lang = lang.split("-")[0]

        if lang in available_languages:
            return lang

    return None


def to_integer(text: str) -> int | None:
    try:
        return int(text)
    except ValueError:
        return None


def choose_language(request: fastapi.Request) -> str:
    from brigid.library.storage import storage

    path = request.url.path

    for language in storage.get_site().allowed_languages:
        if path.startswith(f"/{language}/") or path == f"/{language}":
            return language

    accept_language = request.headers.get("accept-language", "")
    languages = parse_accept_language(accept_language)

    site = storage.get_site()

    if language := select_language(languages, site.allowed_languages):
        return language

    return site.default_language


def translate_seo(language: str, text_id: str) -> str:
    from brigid.library.storage import storage

    site = storage.get_site()

    if text_id in site.languages[language].seo_translations:
        return site.languages[language].seo_translations[text_id]

    if language not in translations:
        language = site.default_language

    if language not in translations:
        language = "en"

    if text_id in translations[language]:
        return translations[language][text_id]

    return text_id


def construct_index_title(  # noqa: CCR001
    language: str,
    title: str,
    page: int,
    tags_required: Iterable[str],
    tags_excluded: Iterable[str],
) -> str:

    if page == 1 and not tags_required and not tags_excluded:
        return title

    title_parts = []

    tags_required = list(tags_required)
    tags_excluded = list(tags_excluded)
    category = None

    if tags_required:
        category = tags_required[0]
        tags_required = tags_required[1:]

    if category:
        title_parts.append(category.capitalize())

    if page > 1:
        if category:
            title_parts.append(", ")
            title_parts.append(translate_seo(language, "page"))
        else:
            title_parts.append(translate_seo(language, "page").capitalize())

        title_parts.append(" ")
        title_parts.append(str(page))

    if tags_required:
        title_parts.append(" ")
        title_parts.append(translate_seo(language, "with_tags"))
        title_parts.append(" ")
        title_parts.append(", ".join([f"[{tag}]" for tag in tags_required]))

    if tags_excluded:
        title_parts.append(" ")
        title_parts.append(translate_seo(language, "without_tags"))
        title_parts.append(" ")
        title_parts.append(", ".join([f"[{tag}]" for tag in tags_excluded]))

    title_parts.append(" | ")
    title_parts.append(title)

    return "".join(title_parts)


def construct_index_description(  # noqa: CCR001
    language: str,
    subtitle: str,
    page: int,
    tags_required: Iterable[str],
    tags_excluded: Iterable[str],
) -> str:

    if page == 1 and not tags_required and not tags_excluded:
        return subtitle

    title_parts = []

    tags_required = list(tags_required)
    tags_excluded = list(tags_excluded)
    category = None

    if tags_required:
        category = tags_required[0]
        tags_required = tags_required[1:]

    if category:
        title_parts.append(category.capitalize())

    if page > 1:
        if category:
            title_parts.append(", ")
            title_parts.append(translate_seo(language, "page"))
        else:
            title_parts.append(translate_seo(language, "page").capitalize())

        title_parts.append(" ")
        title_parts.append(str(page))

    if tags_required:
        title_parts.append(" ")
        title_parts.append(translate_seo(language, "with_tags"))
        title_parts.append(" ")
        title_parts.append(", ".join([f"[{tag}]" for tag in tags_required]))

    if tags_excluded:
        title_parts.append(" ")
        title_parts.append(translate_seo(language, "without_tags"))
        title_parts.append(" ")
        title_parts.append(", ".join([f"[{tag}]" for tag in tags_excluded]))

    title_parts.append(". ")
    title_parts.append(subtitle.capitalize())
    title_parts.append(".")

    return "".join(title_parts)



from collections import Counter
from itertools import chain

from fastapi.responses import HTMLResponse
from feedgenerator import Atom1Feed

from brigid.api.utils import construct_index_description, construct_index_title, to_integer
from brigid.core import errors
from brigid.domain.urls import UrlsFeedsAtom, UrlsPost, UrlsTags
from brigid.library.similarity import get_similar_pages
from brigid.library.storage import storage
from brigid.theme.entities import MetaInfo
from brigid.theme.jinjaglobals import render_page_intro
from brigid.theme.templates import render


def render_index(language: str, raw_tags: str) -> HTMLResponse:  # noqa: CCR001, CFQ001
    site = storage.get_site()

    if language not in site.allowed_languages:
        raise errors.PageNotFound()

    required = set()
    excluded = set()
    page_number = 1

    parsed_tags = raw_tags.split("/")

    parsed_tags = [tag for tag in parsed_tags if tag]

    if parsed_tags:
        parsed_page_number = to_integer(parsed_tags[-1])

        if parsed_page_number is not None and parsed_page_number > 0:
            page_number = parsed_page_number
            parsed_tags.pop()

    for tag in parsed_tags:
        if tag.startswith("-"):
            excluded.add(tag[1:])
        else:
            required.add(tag)

    for tag in chain(required, excluded):
        if tag not in site.languages[language].tags_translations:
            # TODO: maybe, in the future it is better to make smart redirect and message user about it
            raise errors.PageNotFound()

    site = storage.get_site()

    all_pages = storage.last_pages(language=language, require_tags=required, exclude_tags=excluded)

    tags_count: Counter[str] = Counter()

    for page in all_pages:
        tags_count.update(page.tags)

    pages = all_pages[site.posts_per_page * (page_number - 1) : site.posts_per_page * page_number]

    filter_state = UrlsTags(
        language=language,
        page=page_number,
        required_tags=required,
        excluded_tags=excluded,
    )

    translated_tags_required = [site.languages[language].tags_translations[tag] for tag in required]
    translated_tags_required.sort()

    translated_tags_excluded = [site.languages[language].tags_translations[tag] for tag in excluded]
    translated_tags_excluded.sort()

    seo_title = construct_index_title(
        language=language,
        title=site.languages[language].title,
        page=page_number,
        tags_required=translated_tags_required,
        tags_excluded=translated_tags_excluded,
    )

    seo_description = construct_index_description(
        language=language,
        subtitle=site.languages[language].subtitle,
        page=page_number,
        tags_required=translated_tags_required,
        tags_excluded=translated_tags_excluded,
    )

    meta_info = MetaInfo(
        language=language,
        # TODO: construct language urls for filters
        #       it is complicated, because
        #       - we need to reset page number to 1 when changing language
        #       - we need to reapply filters to pages to calculate total pages
        allowed_languages=[],
        title=seo_title,
        description=seo_description,
        author=site.languages[language].author,
        tags=translated_tags_required,
        published_at=None,
        seo_image_url=None,
    )

    # TODO: template names should have prefixes
    content = render(
        "./blog_index.html.j2",
        {
            "language": language,
            "meta_info": meta_info,
            "site": site,
            "pages": pages,
            "current_url": filter_state,
            "article": None,
            "pages_found": len(all_pages),
            "tags_count": tags_count,
        },
    )

    return HTMLResponse(content=content)


def render_page(language: str, article_slug: str, status_code: int = 200) -> HTMLResponse:

    site = storage.get_site()

    if language not in site.allowed_languages:
        raise errors.PageNotFound()

    if not storage.has_article(slug=article_slug):
        raise errors.PageNotFound()

    article = storage.get_article(slug=article_slug)

    if language not in article.pages:
        raise errors.PageNotFound()

    page = storage.get_page(id=article.pages[language])

    similar_pages = get_similar_pages(language=language, original_page=page, number=site.posts_in_similar)

    # TODO: default template for each page type
    template = page.template or site.default_page_template

    post_url = UrlsPost(language=page.language, slug=article.slug)

    seo_image_url = None

    if page.seo_image:
        seo_image_url = post_url.file_url(page.seo_image)

    meta_info = MetaInfo(
        language=language,
        allowed_languages=[language for language in site.allowed_languages if language in article.pages],
        title=page.title,
        description=page.description,
        author=site.languages[language].author,
        tags=[site.languages[language].tags_translations[tag] for tag in page.tags],
        published_at=page.published_at,
        seo_image_url=seo_image_url,
    )

    content = render(
        template,
        {
            "language": language,
            "meta_info": meta_info,
            "site": storage.get_site(),
            "article": article,
            "similar_pages": similar_pages,
            "page": page,
            "current_url": post_url,
        },
    )

    return HTMLResponse(content=content, status_code=status_code)


# TODO: cache?
def render_atom_feed(language: str) -> HTMLResponse:
    site = storage.get_site()

    if language not in site.allowed_languages:
        raise errors.PageNotFound()

    author = site.languages[language].author

    feed = Atom1Feed(
        title=site.languages[language].title,
        # TODO: what is to use: subtitle or description?
        subtitle=site.languages[language].subtitle,
        description=site.languages[language].subtitle,
        link=site.url,
        language=language,
        feed_url=UrlsFeedsAtom(language=language).url(),
        author_name=author,
        # 'image' TODO: favicon?
        # 'author_email': to_unicode(author_email),  # noqa: E800
        # 'author_link': iri_to_uri(author_link),    # noqa: E800
    )

    all_pages = storage.last_pages(language=language)

    for page in all_pages[: site.posts_in_feed]:
        article = storage.get_article(id=page.article_id)

        post_url = UrlsPost(language=language, slug=article.slug)

        feed.add_item(
            # TODO: do we need to render title as markdown?
            title=page.title,
            link=post_url.url(),
            description=render_page_intro(page),
            author_name=author,
            pubdate=page.published_at,
            # TODO: now unique_id is constructed automatically like tag URI, which is expected by ATOM spec
            #       check if it is ok
        )

    xml = feed.writeString("utf-8")

    return HTMLResponse(content=xml, media_type=feed.mime_type)



from typing import Callable

from brigid.library.storage import storage
from brigid.validation.entities import Error


def required_article(slug: str) -> Callable[[], list[Error]]:
    def validator() -> list[Error]:

        if not storage.has_article(slug=slug):
            return [Error(filepath=None, message=f"{slug} article not found")]

        article = storage.get_article(slug=slug)

        site = storage.get_site()

        allowed_languages = set(site.allowed_languages)
        article_languages = set(article.pages)

        missing_languages = allowed_languages - article_languages

        if missing_languages:
            return [Error(filepath=article.path, message=f"Missing languages for article: {missing_languages}")]

        return []

    return validator
